/*
 * Python extension for Field Reports 
 * 
 * Copyright (c) Field Works, LLC
 */
#include "Python.h"
#if (PY_MAJOR_VERSION >= 2 && PY_MINOR_VERSION >= 6) || PY_MAJOR_VERSION >= 3
#include "bytesobject.h"
#else
#define PyBytes_Check PyString_Check
#define PyBytes_AsString PyString_AsString
#define PyBytes_FromStringAndSize PyString_FromStringAndSize
#endif

#include <reports.h>

struct module_state
{
    PyObject *REPORTS_Error;
    PyObject *INPUT_Error;
    PyObject *RENDER_Error;
    PyObject *PDF_Error;
    PyObject *AUTH_Error;
    PyObject *RUNTIME_Error;
    PyObject *SIG_Error;
};

#if PY_MAJOR_VERSION >= 3
#define GETSTATE(m) ((struct module_state *)PyModule_GetState(m))
#else
#define GETSTATE(m) (&_state)
static struct module_state _state;
#endif

static int py_type(PyObject *obj)
{
    if (obj == Py_None)
    {
        return RP_TYPE_NULL;
    }
    else if (PyBool_Check(obj))
    {
        return RP_TYPE_BOOL;
#if PY_MAJOR_VERSION < 3
    }
    else if (PyInt_Check(obj))
    {
        return RP_TYPE_INT;
#endif
    }
    else if (PyLong_Check(obj))
    {
        return RP_TYPE_INT;
    }
    else if (PyFloat_Check(obj))
    {
        return RP_TYPE_FLOAT;
    }
    else if (PyUnicode_Check(obj))
    {
        return RP_TYPE_STRING;
    }
    else if (PyBytes_Check(obj))
    {
        return RP_TYPE_STRING;
    }
    else if (PyList_Check(obj))
    {
        return RP_TYPE_ARRAY;
    }
    else if (PyTuple_Check(obj))
    {
        return RP_TYPE_ARRAY;
    }
    else if (PyDict_Check(obj))
    {
        return RP_TYPE_OBJECT;
    }
    else
    {
        return RP_TYPE_OTHER;
    }
}

static void py_release(PyObject *opt)
{
    if (opt)
    {
        Py_DECREF(opt);
    }
}

static int py_val_bool(PyObject *obj, PyObject **opt)
{
    *opt = NULL;
    return (obj == Py_True);
}

static long py_val_int(PyObject *obj, PyObject **opt)
{
    *opt = NULL;
#if PY_MAJOR_VERSION < 3
    if (PyInt_Check(obj))
    {
        return PyInt_AsLong(obj);
    }
    else
    {
        return PyLong_AsLong(obj);
    }
#else
    return PyLong_AsLong(obj);
#endif
}

static double py_val_float(PyObject *obj, PyObject **opt)
{
    *opt = NULL;
    return PyFloat_AsDouble(obj);
}

static char *py_val_string(PyObject *obj, PyObject **opt)
{
    *opt = NULL;
    if (PyUnicode_Check(obj))
    {
        *opt = PyUnicode_AsUTF8String(obj);
        return PyBytes_AsString(*opt);
    }
    else if (PyBytes_Check(obj))
    {
        return PyBytes_AsString(obj);
    }
    else
    {
        return NULL;
    }
}

static Py_ssize_t py_array_head(PyObject *obj, PyObject **opt)
{
    *opt = NULL;
    return 0;
}

static int py_array_next(PyObject *obj, Py_ssize_t *pos, PyObject **entry, PyObject **opt)
{
    *opt = NULL;
    if (PyList_Check(obj))
    {
        if (*pos < PyList_Size(obj))
        {
            *entry = PyList_GetItem(obj, *pos);
            ++(*pos);
            return 1;
        }
        else
        {
            *entry = NULL;
            return 0;
        }
    }
    else if (PyTuple_Check(obj))
    {
        if (*pos < PyTuple_Size(obj))
        {
            *entry = PyTuple_GetItem(obj, *pos);
            ++(*pos);
            return 1;
        }
        else
        {
            *entry = NULL;
            return 0;
        }
    }
    else
    {
        *entry = NULL;
        return 0;
    }
}

static Py_ssize_t py_object_head(PyObject *obj, PyObject **opt)
{
    *opt = NULL;
    return 0;
}

static int py_object_next(PyObject *obj, Py_ssize_t *pos, char **key, PyObject **entry, PyObject **opt)
{
    PyObject *py_key;
    int result = PyDict_Next(obj, pos, &py_key, entry);
    *opt = NULL;
    if (result)
    {
        if (py_type(py_key) == RP_TYPE_STRING)
        {
            *key = py_val_string(py_key, opt);
        }
        else
        {
            PyObject *skey = PyObject_Str(py_key);
            *key = py_val_string(skey, opt);
            if (*opt)
            {
                Py_DECREF(skey);
            }
        }
    }
    return result;
}

rp_callback callbacks = {
    (t_fn_type)py_type,
    (t_fn_val_bool)py_val_bool,
    (t_fn_val_int)py_val_int,
    (t_fn_val_float)py_val_float,
    (t_fn_val_string)py_val_string,
    (t_fn_array_head)py_array_head,
    (t_fn_array_next)py_array_next,
    (t_fn_object_head)py_object_head,
    (t_fn_object_next)py_object_next,
    (t_fn_release)py_release,
};

static PyObject *get_py_bytes(caml_value *v)
{
    PyObject *py_result = NULL;
    int len = rp_get_string_val(v, NULL);
    char *buf = malloc(len + 1);
    if (buf)
    {
        rp_get_string_val(v, buf);
        buf[len] = '\0';
        py_result = PyBytes_FromStringAndSize(buf, len);
        free(buf);
    }
    return py_result;
}

static PyObject *get_py_string(caml_value *v)
{
    PyObject *py_result = NULL;
    int len = rp_get_string_val(v, NULL);
    char *buf = malloc(len + 1);
    if (buf)
    {
        rp_get_string_val(v, buf);
        buf[len] = '\0';
        py_result = PyUnicode_DecodeUTF8(buf, len, NULL);
        free(buf);
    }
    return py_result;
}

static PyObject *get_py_error(struct module_state *st, int no)
{
    switch (no)
    {
    case RP_REPORTS_ERROR:
        return st->REPORTS_Error;
    case RP_INPUT_ERROR:
        return st->INPUT_Error;
    case RP_RENDER_ERROR:
        return st->RENDER_Error;
    case RP_PDF_ERROR:
        return st->PDF_Error;
    case RP_AUTH_ERROR:
        return st->AUTH_Error;
    case RP_SIG_ERROR:
        return st->SIG_Error;
    default:
        return st->RUNTIME_Error;
    }
}

static void set_error_msg(PyObject *self, caml_value *exn)
{
    char msg[RP_MAX_ERR_MSG_LEN];
    int err;
    struct module_state *st = GETSTATE(self);

    rp_get_error(exn, &err, msg);
    PyErr_SetString(get_py_error(st, err), msg);
}

static PyObject *py_version(PyObject *self, PyObject *args)
{
    PyObject *py_result = NULL;
    caml_value *caml_result = NULL;

    // check num arguments
    if (PyTuple_Check(args) && PyTuple_Size(args) != 0)
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // call OCaml function
    caml_result = rp_version();
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

    // set return value
    py_result = get_py_string(caml_result);

exit:
    if (caml_result) rp_free(caml_result);

    return py_result;
}

static PyObject *py_set_defaults(PyObject *self, PyObject *args, PyObject *keywds)
{
    static char *kwlist[] = {"param", "debug", NULL};
    PyObject *py_param = NULL;
    PyObject *py_result = NULL;
    caml_value *caml_param = NULL;
    caml_value *caml_result = NULL;
    int debug = 0;

    // get arguments
    if (!PyArg_ParseTupleAndKeywords(args, keywds, "O|i", kwlist, &py_param, &debug))
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // call OCaml function
    caml_param = rp_encode_json(&callbacks, py_param);
    if (rp_is_error(caml_param))
    {
        set_error_msg(self, caml_param);
        goto exit;
    }

    caml_result = rp_set_defaults(caml_param);
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

    if (debug)
    {
        // return param as string (for test)
        py_result = get_py_string(caml_result);
    }
    else
    {
        Py_INCREF(Py_None);
        py_result = Py_None;
    }

exit:
    if (caml_param) rp_free(caml_param);
    if (caml_result) rp_free(caml_result);

    return py_result;
}

static PyObject *py_set_log_level(PyObject *self, PyObject *args)
{
    caml_value *caml_result = NULL;
    PyObject *py_level = NULL;
    long level = 0;

    // check num arguments
    if (PyTuple_Check(args) && PyTuple_Size(args) != 1)
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // convert parameter
    py_level = PyTuple_GetItem(args, 0);
    if (PyLong_Check(py_level))
    {
        level = PyLong_AsLong(py_level);
        if (PyErr_Occurred() != NULL)
        {
            goto exit;
        }
#if PY_MAJOR_VERSION < 3
    }
    else if (PyInt_Check(py_level))
    {
        level = PyInt_AsLong(py_level);
        if (PyErr_Occurred() != NULL)
        {
            goto exit;
        }
#endif
    }
    else
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // call OCaml function
    caml_result = rp_set_log_level((int)level);
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

exit:
    if (caml_result) rp_free(caml_result);

    // set return value
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *py_set_log_flag(PyObject *self, PyObject *args)
{
    caml_value *caml_result = NULL;
    PyObject *py_flag = NULL;
    long flag = 0;

    // check num arguments
    if (PyTuple_Check(args) && PyTuple_Size(args) != 1)
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // convert parameter
    py_flag = PyTuple_GetItem(args, 0);
    if (PyLong_Check(py_flag))
    {
        flag = PyLong_AsLong(py_flag);
        if (PyErr_Occurred() != NULL)
        {
            goto exit;
        }
#if PY_MAJOR_VERSION < 3
    }
    else if (PyInt_Check(py_flag))
    {
        flag = PyInt_AsLong(py_flag);
        if (PyErr_Occurred() != NULL)
        {
            goto exit;
        }
#endif
    }
    else
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // call OCaml function
    caml_result = rp_set_log_flag((int)flag);
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

exit:
    if (caml_result) rp_free(caml_result);

    // set return value
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *py_renders(PyObject *self, PyObject *args)
{
    caml_value *caml_param = NULL;
    caml_value *caml_result = NULL;
    PyObject *py_param = NULL;
    PyObject *py_result = NULL;

    // check num arguments
    if (PyTuple_Check(args) && PyTuple_Size(args) != 1)
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // convert parameter
    py_param = PyTuple_GetItem(args, 0);
    caml_param = rp_encode_json(&callbacks, py_param);
    if (rp_is_error(caml_param))
    {
        set_error_msg(self, caml_param);
        goto exit;
    }

    // call OCaml function
    caml_result = rp_pdf_of_json(caml_param);
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

    // set return value
    py_result = get_py_bytes(caml_result);

exit:
    if (caml_param) rp_free(caml_param);
    if (caml_result) rp_free(caml_result);
    return py_result;
}

static PyObject *py_render(PyObject *self, PyObject *args)
{
    caml_value *caml_param = NULL;
    caml_value *caml_result = NULL;
    PyObject *py_fname = NULL;

    // check num arguments
    if (PyTuple_Check(args) && PyTuple_Size(args) != 2)
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // get parameters
    caml_param = rp_encode_json(&callbacks, PyTuple_GetItem(args, 0));
    if (rp_is_error(caml_param))
    {
        set_error_msg(self, caml_param);
        goto exit;
    }
    py_fname = PyTuple_GetItem(args, 1);

    // call OCaml function
    if (PyUnicode_Check(py_fname))
    {
        PyObject *unicode = PyUnicode_AsUTF8String(py_fname);
        caml_result = rp_write_pdf_from_json(caml_param, PyBytes_AsString(unicode));
        Py_DECREF(unicode);
    }
    else if (PyBytes_Check(py_fname))
    {
        caml_result = rp_write_pdf_from_json(caml_param, PyBytes_AsString(py_fname));
    }
    else
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

exit:
    if (caml_param) rp_free(caml_param);
    if (caml_result) rp_free(caml_result);

    // set return value
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *py_signs(PyObject *self, PyObject *args)
{
    caml_value *caml_param = NULL;
    caml_value *caml_result = NULL;
    PyObject *py_param = NULL;
    PyObject *py_result = NULL;

    // check num arguments
    if (PyTuple_Check(args) && PyTuple_Size(args) != 1)
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // convert parameter
    py_param = PyTuple_GetItem(args, 0);
    caml_param = rp_encode_json(&callbacks, py_param);
    if (rp_is_error(caml_param))
    {
        set_error_msg(self, caml_param);
        goto exit;
    }

    // call OCaml function
    caml_result = rp_signedpdf_of_json(caml_param);
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

    // set return value
    py_result = get_py_bytes(caml_result);

exit:
    if (caml_param) rp_free(caml_param);
    if (caml_result) rp_free(caml_result);
    return py_result;
}

static PyObject *py_sign(PyObject *self, PyObject *args)
{
    caml_value *caml_param = NULL;
    caml_value *caml_result = NULL;
    PyObject *py_fname = NULL;

    // check num arguments
    if (PyTuple_Check(args) && PyTuple_Size(args) != 2)
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // get parameters
    caml_param = rp_encode_json(&callbacks, PyTuple_GetItem(args, 0));
    if (rp_is_error(caml_param))
    {
        set_error_msg(self, caml_param);
        goto exit;
    }
    py_fname = PyTuple_GetItem(args, 1);

    // call OCaml function
    if (PyUnicode_Check(py_fname))
    {
        PyObject *unicode = PyUnicode_AsUTF8String(py_fname);
        caml_result = rp_write_signedpdf_from_json(caml_param, PyBytes_AsString(unicode));
        Py_DECREF(unicode);
    }
    else if (PyBytes_Check(py_fname))
    {
        caml_result = rp_write_signedpdf_from_json(caml_param, PyBytes_AsString(py_fname));
    }
    else
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

exit:
    if (caml_param) rp_free(caml_param);
    if (caml_result) rp_free(caml_result);

    // set return value
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *py_validates(PyObject *self, PyObject *args)
{
    caml_value *caml_param = NULL;
    caml_value *caml_result = NULL;
    PyObject *py_param = NULL;
    PyObject *py_result = NULL;

    // check num arguments
    if (PyTuple_Check(args) && PyTuple_Size(args) != 1)
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // convert parameter
    py_param = PyTuple_GetItem(args, 0);
    caml_param = rp_encode_json(&callbacks, py_param);
    if (rp_is_error(caml_param))
    {
        set_error_msg(self, caml_param);
        goto exit;
    }

    // call OCaml function
    caml_result = rp_validatedpdf_of_json(caml_param);
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

    // set return value
    py_result = PyTuple_New(2);
    PyTuple_SetItem(py_result, 0, get_py_string(rp_get_field(caml_result, 0)));
    PyTuple_SetItem(py_result, 1, get_py_bytes(rp_get_field(caml_result, 1)));

exit:
    if (caml_param) rp_free(caml_param);
    if (caml_result) rp_free(caml_result);
    return py_result;
}

static PyObject *py_validate(PyObject *self, PyObject *args)
{
    caml_value *caml_param = NULL;
    caml_value *caml_result = NULL;
    PyObject *py_fname = NULL;
    PyObject *py_result = NULL;

    // check num arguments
    if (PyTuple_Check(args) && PyTuple_Size(args) != 2)
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }

    // get parameters
    caml_param = rp_encode_json(&callbacks, PyTuple_GetItem(args, 0));
    if (rp_is_error(caml_param))
    {
        set_error_msg(self, caml_param);
        goto exit;
    }
    py_fname = PyTuple_GetItem(args, 1);

    // call OCaml function
    if (PyUnicode_Check(py_fname))
    {
        PyObject *unicode = PyUnicode_AsUTF8String(py_fname);
        caml_result = rp_write_validatedpdf_from_json(caml_param, PyBytes_AsString(unicode));
        Py_DECREF(unicode);
    }
    else if (PyBytes_Check(py_fname))
    {
        caml_result = rp_write_validatedpdf_from_json(caml_param, PyBytes_AsString(py_fname));
    }
    else
    {
        struct module_state *st = GETSTATE(self);
        PyErr_SetString(st->REPORTS_Error, "illegal arguments");
        goto exit;
    }
    if (rp_is_error(caml_result))
    {
        set_error_msg(self, caml_result);
        goto exit;
    }

    // set return value
    py_result = get_py_string(caml_result);

exit:
    if (caml_param) rp_free(caml_param);
    if (caml_result) rp_free(caml_result);

    // set return value
    Py_INCREF(Py_None);
    return py_result;
}

static PyMethodDef reports_methods[] = {
    {"version", py_version, METH_VARARGS, "get version number."},
    {"set_defaults", (PyCFunction)py_set_defaults, METH_VARARGS | METH_KEYWORDS,
     "set default rendering parameters."},
    {"set_log_level", py_set_log_level, METH_VARARGS, "set log level."},
    {"set_log_flag", py_set_log_flag, METH_VARARGS, "set log flag."},
    {"renders", py_renders, METH_VARARGS, "create PDF string from json object or string."},
    {"render", py_render, METH_VARARGS, "write PDF file from json object or string."},
    {"signs", py_signs, METH_VARARGS, "create signed PDF string from json object or string."},
    {"sign", py_sign, METH_VARARGS, "write signed PDF file from json object or string."},
    {"validates", py_validates, METH_VARARGS, "create validated PDF string from json object or string."},
    {"validate", py_validate, METH_VARARGS, "write validated PDF file from json object or string."},
    {NULL, NULL, 0, NULL} /* Sentinel */
};

PyDoc_STRVAR(reports_doc, "Field Reports: PDF report genaration tool.");

#if PY_MAJOR_VERSION >= 3

static int reports_traverse(PyObject *m, visitproc visit, void *arg)
{
    Py_VISIT(GETSTATE(m)->REPORTS_Error);
    Py_VISIT(GETSTATE(m)->INPUT_Error);
    Py_VISIT(GETSTATE(m)->RENDER_Error);
    Py_VISIT(GETSTATE(m)->PDF_Error);
    Py_VISIT(GETSTATE(m)->AUTH_Error);
    Py_VISIT(GETSTATE(m)->RUNTIME_Error);
    Py_VISIT(GETSTATE(m)->SIG_Error);
    return 0;
}

static int reports_clear(PyObject *m)
{
    Py_CLEAR(GETSTATE(m)->REPORTS_Error);
    Py_CLEAR(GETSTATE(m)->INPUT_Error);
    Py_CLEAR(GETSTATE(m)->RENDER_Error);
    Py_CLEAR(GETSTATE(m)->PDF_Error);
    Py_CLEAR(GETSTATE(m)->AUTH_Error);
    Py_CLEAR(GETSTATE(m)->RUNTIME_Error);
    Py_CLEAR(GETSTATE(m)->SIG_Error);
    return 0;
}

static struct PyModuleDef moduledef = {
    PyModuleDef_HEAD_INIT,
    "reports",
    reports_doc,
    sizeof(struct module_state),
    reports_methods,
    NULL,
    reports_traverse,
    reports_clear,
    NULL};

#define INIT_ERROR_RETURN return NULL

#else

#define INIT_ERROR_RETURN return

#endif

PyMODINIT_FUNC
#if PY_MAJOR_VERSION >= 3
PyInit_reports(void)
#else
initreports(void)
#endif
{
#if PY_MAJOR_VERSION >= 3
    PyObject *m = PyModule_Create(&moduledef);
#else
    PyObject *m = Py_InitModule3("reports", reports_methods, reports_doc);
#endif
    struct module_state *st = GETSTATE(m);
    if (m == NULL)
        INIT_ERROR_RETURN;

    st->REPORTS_Error = PyErr_NewException("reports.Error", NULL, NULL);
    if (st->REPORTS_Error == NULL)
    {
        Py_DECREF(m);
        INIT_ERROR_RETURN;
    }
    Py_INCREF(st->REPORTS_Error);
    PyModule_AddObject(m, "Error", st->REPORTS_Error);

    st->INPUT_Error = PyErr_NewException("reports.InputError", NULL, NULL);
    if (st->INPUT_Error == NULL)
    {
        Py_DECREF(m);
        INIT_ERROR_RETURN;
    }
    Py_INCREF(st->INPUT_Error);
    PyModule_AddObject(m, "InputError", st->INPUT_Error);

    st->RENDER_Error = PyErr_NewException("reports.RenderError", NULL, NULL);
    if (st->RENDER_Error == NULL)
    {
        Py_DECREF(m);
        INIT_ERROR_RETURN;
    }
    Py_INCREF(st->RENDER_Error);
    PyModule_AddObject(m, "RenderError", st->RENDER_Error);

    st->PDF_Error = PyErr_NewException("reports.PDFError", NULL, NULL);
    if (st->PDF_Error == NULL)
    {
        Py_DECREF(m);
        INIT_ERROR_RETURN;
    }
    Py_INCREF(st->PDF_Error);
    PyModule_AddObject(m, "PDFError", st->PDF_Error);

    st->AUTH_Error = PyErr_NewException("reports.AuthError", NULL, NULL);
    if (st->AUTH_Error == NULL)
    {
        Py_DECREF(m);
        INIT_ERROR_RETURN;
    }
    Py_INCREF(st->AUTH_Error);
    PyModule_AddObject(m, "AuthError", st->AUTH_Error);

    st->RUNTIME_Error = PyErr_NewException("reports.RuntimeError", NULL, NULL);
    if (st->RUNTIME_Error == NULL)
    {
        Py_DECREF(m);
        INIT_ERROR_RETURN;
    }
    Py_INCREF(st->RUNTIME_Error);
    PyModule_AddObject(m, "RuntimeError", st->RUNTIME_Error);

    st->SIG_Error = PyErr_NewException("reports.SigError", NULL, NULL);
    if (st->SIG_Error == NULL)
    {
        Py_DECREF(m);
        INIT_ERROR_RETURN;
    }
    Py_INCREF(st->SIG_Error);
    PyModule_AddObject(m, "SigError", st->SIG_Error);

    rp_init();

#if PY_MAJOR_VERSION >= 3
    return m;
#endif
}
